package main

import (
	"context"
	"fmt"
	"io"
	"os"
	"os/signal"
	"path"
	"path/filepath"
	"strings"
	"sync"
	"syscall"
	"time"

	_ "unsafe"

	// "chainmaker.org/chainmaker/bloom"
	"chainmaker.org/chainmaker/localconf/v2"
	"chainmaker.org/chainmaker/logger/v2"
	storePb "chainmaker.org/chainmaker/pb-go/v2/store"
	"chainmaker.org/chainmaker/protocol/v2"
	"chainmaker.org/chainmaker/store/v2"
	"chainmaker.org/chainmaker/store/v2/blockdb/blockfiledb"
	"chainmaker.org/chainmaker/store/v2/bloom"
	"chainmaker.org/chainmaker/store/v2/conf"
	"chainmaker.org/chainmaker/store/v2/dbprovider"
	"chainmaker.org/chainmaker/store/v2/serialization"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
	"github.com/syndtr/goleveldb/leveldb"
	"github.com/syndtr/goleveldb/leveldb/opt"
)

const (
	blockFilePath = "bfdb"
	walLogPath    = "wal"
)

var (
	configFile       string
	itemsN           int64
	fp               float64
	chainMakerConfig *localconf.CMConfig
	workDir          string
	currentDir       string
	logLevel         string
	threadNum        int32
	dumpCycle        int32 = 10000
	workOneThread    int32
	// minWorkOneThread int32 = 100
)

func init() {
	defaultConf := logger.GetDefaultLogNodeConfig()
	defaultConf.LogLevelDefault = "INFO"
	logger.SetLogConfig(&logger.LogConfig{
		SystemLog: defaultConf,
	})
}

func main() {
	cmd := &cobra.Command{
		Use:   "./cmhdbfb",
		Short: "bloom load tool for chainmaker",
		// RunE: func(cmd *cobra.Command, _ []string) error {
		// 	defaultConf := logger.GetDefaultLogNodeConfig()
		// 	defaultConf.LogLevelDefault = "INFO"
		// 	logger.SetLogConfig(&logger.LogConfig{
		// 		SystemLog: defaultConf,
		// 	})
		// 	return nil
		// },
	}
	cmd.AddCommand(loadDataCmd())
	cmd.AddCommand(verifyKeyCmd())
	cmd.AddCommand(historyKeyRead())
	cmd.Execute()
}
func setLog(logLevel string) {
	defaultConf := logger.GetDefaultLogNodeConfig()
	defaultConf.LogLevelDefault = logLevel
	logger.SetLogConfig(&logger.LogConfig{
		SystemLog: defaultConf,
	})
}
func loadDataCmd() *cobra.Command {
	cmd := &cobra.Command{
		Use:   "load",
		Short: "Load write sets of block which from file db to bloom filter",
		RunE: func(cmd *cobra.Command, _ []string) error {
			if itemsN <= 0 || fp <= 0 {
				return fmt.Errorf("invalid bloom filter params")
			}
			fmt.Printf("thread[%d], workPerThread[%d]\n", threadNum, workOneThread)
			if threadNum <= 0 {
				threadNum = 1
			}
			setLog(logLevel)
			currentDir, _ = os.Getwd()
			loadChainMakerConfig(cmd)
			if workDir != "" {
				if err := os.Chdir(workDir); err != nil {
					return err
				}
			}
			ctx := setupSignalContext()
			run(ctx)
			return nil
		},
	}
	flags := &pflag.FlagSet{}
	flags.StringVarP(&configFile, "config", "c", "chainmaker.yml", "config file of chainmaker")
	flags.Int64VarP(&itemsN, "items", "n", 100000000, "bloom filter keys capacity")
	flags.Float64VarP(&fp, "fpr", "r", 0.01, "bloom filter false positive rate")
	flags.StringVarP(&workDir, "workdir", "d", "", "work dir used to switch the execution directory to correctly access the relative path in config file")
	flags.StringVarP(&logLevel, "log", "l", "INFO", "logger level output")
	flags.Int32VarP(&threadNum, "thread", "t", 1, "thread num, if <= 0, will use one")
	flags.Int32VarP(&workOneThread, "workperthread", "w", 10000, "workload per thread, if <=0, distribute workload automatically")

	cmd.Flags().AddFlag(flags.Lookup("config"))
	cmd.Flags().AddFlag(flags.Lookup("items"))
	cmd.Flags().AddFlag(flags.Lookup("fpr"))
	cmd.Flags().AddFlag(flags.Lookup("workdir"))
	cmd.Flags().AddFlag(flags.Lookup("log"))
	cmd.Flags().AddFlag(flags.Lookup("thread"))
	cmd.Flags().AddFlag(flags.Lookup("workperthread"))
	// cmd.MarkFlagRequired("config")
	// cmd.MarkFlagRequired("items")
	// cmd.MarkFlagRequired("fpr")
	return cmd
}

func verifyKeyCmd() *cobra.Command {
	var fileP string
	cmd := &cobra.Command{
		Use:   "verify",
		Short: "Verify key command",
		Long:  "Verify key command",
		PreRun: func(cmd *cobra.Command, args []string) {

		},
		RunE: func(cmd *cobra.Command, _ []string) error {
			logger := logger.GetLoggerByChain("[BloomLoader]", "test")
			bFilter, err := bloom.NewBlockBloomFilter(10000000, 0.01,
				bloom.WithLogger(logger),
				bloom.WithDumpFile(path.Join(fileP, "bloom.dump")),
				bloom.WithContractKeyGenerator(func(contractName string, key []byte) []byte {
					buf := make([]byte, 0, len(contractName)+len(key)+1)
					return append(append(append(buf, []byte(contractName)...), '#'), key...)
				}),
			)
			if err != nil {
				return err
			}
			var (
				contractName string
				key          string
			)
			fmt.Println("please input {contractName} {key}")
			for {
				contractName = ""
				key = ""
				_, err = fmt.Scanf("%s %s", &contractName, &key)
				if err != nil {
					if err == io.EOF {
						fmt.Println("read eof exit")
						break
					} else {
						logger.Errorf("read error: %s", err)
						continue
					}
				}
				logger.Infof("contractName: %s, key: %s, exist: %t", contractName, key, bFilter.Has(contractName, []byte(key)))
			}
			return nil
		},
	}
	flags := &pflag.FlagSet{}
	flags.StringVarP(&fileP, "file", "f", "", "bloom file for load data")
	cmd.Flags().AddFlag(flags.Lookup("file"))
	cmd.MarkFlagRequired("file")
	return cmd
}

func historyKeyRead() *cobra.Command {
	var historyP string
	cmd := &cobra.Command{
		Use:   "history",
		Short: "Read keys from history db",
		RunE: func(cmd *cobra.Command, _ []string) error {
			if _, err := os.Stat(historyP); err != nil {
				return err
			}
			setLog(logLevel)
			logger := logger.GetLoggerByChain("[BloomLoader]", "test")
			db, err := openLevelDB(historyP)
			if err != nil {
				return err
			}
			iterator := db.NewIterator(nil, nil)
			//history db 存储格式 k{contractName}#{key}#{block_height}#{tx_index}${tx_id}
			for iterator.Next() {
				logger.Debugf("key: %s", string(iterator.Key()))
				key := strings.TrimLeft(string(iterator.Key()), "k")
				if strings.Count(key, "#") < 3 {
					continue
				}
				key = key[:strings.LastIndexByte(key[:strings.LastIndexByte(key, '#')], '#')]
				strs := strings.SplitN(key, "#", 2)
				if len(strs) < 2 {
					// fmt.Println(string(iterator.Key()))
					continue
				}
				fmt.Printf("%s %s\n", strs[0], strs[1])
			}
			return nil
		},
	}
	flags := &pflag.FlagSet{}
	flags.StringVarP(&historyP, "db_path", "p", "", "history db path")
	flags.StringVarP(&logLevel, "log", "l", "INFO", "logger level output")
	cmd.Flags().AddFlag(flags.Lookup("db_path"))
	cmd.Flags().AddFlag(flags.Lookup("log"))
	cmd.MarkFlagRequired("db_path")
	return cmd
}

func loadChainMakerConfig(cmd *cobra.Command) {
	var err error
	chainMakerConfig, err = initLocalConfig(cmd)
	if err != nil {
		fmt.Println(err)
		os.Exit(0)
	}
}

func initLocalConfig(cmd *cobra.Command) (*localconf.CMConfig, error) {
	// 0. load env
	cmViper := viper.New()
	err := cmViper.BindPFlags(cmd.PersistentFlags())
	if err != nil {
		return nil, err
	}
	// 1. load the path of the config files
	ymlFile := configFile
	if !filepath.IsAbs(ymlFile) {
		ymlFile, _ = filepath.Abs(ymlFile)
		configFile = ymlFile
	}

	// 2. load the config file
	cmViper.SetConfigFile(ymlFile)
	if err := cmViper.ReadInConfig(); err != nil {
		return nil, err
	}

	// 3. create new CMConfig instance
	cmConfig := &localconf.CMConfig{} //&CMConfig{}
	if err := cmViper.Unmarshal(cmConfig); err != nil {
		return nil, err
	}
	return cmConfig, nil
}

func run(ctx context.Context) error {
	storageConf, err := conf.NewStorageConfig(chainMakerConfig.StorageConfig)
	if err != nil {
		return err
	}
	if storageConf.DisableBlockFileDb {
		return fmt.Errorf("just support enable block file db")
	}
	wg := sync.WaitGroup{}
	for _, chain := range chainMakerConfig.BlockChainConfig {
		loader, err := newBlockBloomLoader(chain.ChainId, storageConf)
		if err != nil {
			fmt.Printf("newBlockBloomLoader for chain[%s] error %s\n", chain.ChainId, err.Error())
			return err
		}
		wg.Add(1)
		go func() {
			defer wg.Done()
			loader.LoadBloom(ctx.Done())
			loader.Close()
		}()
	}
	wg.Wait()
	return nil
}

func setupSignalContext() context.Context {
	shutdownHandler := make(chan os.Signal, 2)

	ctx, cancel := context.WithCancel(context.Background())
	signal.Notify(shutdownHandler, os.Interrupt, syscall.SIGTERM, syscall.SIGINT)
	go func() {
		<-shutdownHandler
		fmt.Println("handle pre-closure then shut down")
		fmt.Println("如果是加载数据操作，保存数据完成后其会自行退出,请等待...")
		fmt.Println("如果再次触发会强制退出有可能导致数据错误。。。。。。")
		cancel()
		//再次触发直接退出
		<-shutdownHandler
		fmt.Println("Shutting down...")
		os.Exit(1)
	}()
	return ctx
}

func newKvDBHandler(chainID string, dbConfig *conf.DbConfig, logger protocol.Logger) (protocol.DBHandle, error) {
	factory := dbprovider.NewDBFactory()
	return factory.NewKvDB(chainID,
		dbConfig.Provider,
		store.StoreBlockDBDir,
		dbConfig.GetDbConfig(),
		logger,
		nil,
	)
}

func newFileLogger(chainID string, logger protocol.Logger,
	storeConfig *conf.StorageConfig) (*blockfiledb.BlockFile, error) {
	opts := blockfiledb.DefaultOptions
	opts.NoCopy = true
	opts.NoSync = storeConfig.LogDBSegmentAsync
	if storeConfig.LogDBSegmentSize > 64 { // LogDBSegmentSize default is 64MB
		opts.SegmentSize = storeConfig.LogDBSegmentSize * 1024 * 1024
	}
	if storeConfig.DisableLogDBMmap {
		opts.UseMmap = false
	}

	if storeConfig.ReadBFDBTimeOut > 0 {
		opts.ReadTimeOut = storeConfig.ReadBFDBTimeOut
	}

	bfdbPath := filepath.Join(storeConfig.StorePath, chainID, blockFilePath)
	tmpbfdbPath := ""
	if len(storeConfig.BlockStoreTmpPath) > 0 {
		tmpbfdbPath = filepath.Join(storeConfig.BlockStoreTmpPath, chainID, blockFilePath)
	}
	fmt.Println(bfdbPath, tmpbfdbPath)
	bfdb, err := blockfiledb.Open(bfdbPath, tmpbfdbPath, opts, logger)
	if err != nil {
		panic(fmt.Sprintf("open block file db failed, path:%s, error:%s", bfdbPath, err))
	}
	return bfdb, nil
}

func newBloomFilter(chainID string, logger protocol.Logger) (*bloom.BlockBloomFilter, error) {
	fpath := path.Join(currentDir, chainID)
	os.Mkdir(fpath, os.ModePerm)
	opts := []bloom.Option{
		bloom.WithLogger(logger),
		bloom.WithDumpFile(path.Join(fpath, "bloom.dump")),
		bloom.WithContractKeyGenerator(func(contractName string, key []byte) []byte {
			buf := make([]byte, 0, len(contractName)+len(key)+1)
			return append(append(append(buf, []byte(contractName)...), '#'), key...)
		}),
	}
	if threadNum == 1 {
		opts = append(opts, bloom.WithDumpCycle(int(dumpCycle)), bloom.WithDumpAsync())
	} else {
		opts = append(opts, bloom.WithDumpCycle(0))
	}
	return bloom.NewBlockBloomFilter(uint(itemsN), fp, opts...)
}

type blockBloomLoader struct {
	chainID     string
	blockFileDB *blockfiledb.BlockFile
	storeConfig *conf.StorageConfig
	logger      protocol.Logger
	filter      *bloom.BlockBloomFilter
}

func newBlockBloomLoader(chainID string, storageConf *conf.StorageConfig) (*blockBloomLoader, error) {
	logger := logger.GetLoggerByChain("[BloomLoader]", chainID)
	if _, err := os.Stat(storageConf.StorePath); err != nil {
		logger.Error(err)
		return nil, err
	}

	fileLogger, err := newFileLogger(chainID, logger, storageConf)
	if err != nil {
		return nil, err
	}
	filter, err := newBloomFilter(chainID, logger)
	if err != nil {
		return nil, err
	}
	return &blockBloomLoader{
		chainID:     chainID,
		blockFileDB: fileLogger,
		storeConfig: storageConf,
		logger:      logger,
		filter:      filter,
	}, nil
}

func (bl *blockBloomLoader) GetBlockWithRWSets(height uint64) (*storePb.BlockWithRWSet, error) {
	data, _, _, _, err := bl.blockFileDB.ReadLastSegSection(height+1, true)
	if err != nil {
		return nil, err
	}
	return serialization.DeserializeBlock(data)
}
func (bl *blockBloomLoader) IterateBlockFrom(height uint64) (*blockIterator, error) {
	datas, err := bl.blockFileDB.ReadAllSegSections(height + 1)
	if err != nil {
		return nil, err
	}
	return &blockIterator{
		data: datas,
		from: height,
	}, nil
}

func (bl *blockBloomLoader) getLastHeight() (uint64, error) {
	h, err := bl.blockFileDB.LastIndex()
	if err != nil {
		return 0, err
	}
	return h - 1, nil
}

func (bl *blockBloomLoader) LoadBloom(stopCh <-chan struct{}) {
	h := bl.filter.LastDumpHeight()
	//dump高度为0时，需要留意，1. 没有dump 2.dump到了0个区块，为了保险起见，对于0的，继续从0恢复
	if h > 0 {
		h++
	}

	getLastHeight := func() uint64 {
		retry := 10
		for {
			select {
			case <-stopCh:
				break
			default:
			}
			lastHeight, err := bl.getLastHeight()
			if err != nil {
				bl.logger.Errorf("get last block height error: %s", err)
				retry--
				if retry == 0 {
					bl.logger.Errorf("retry %d times, but still failed", 10)
					break
				}
				time.Sleep(20 * time.Millisecond)
				continue
			}
			return lastHeight
		}
		return 0
	}

	for {
		select {
		case <-stopCh:
			return
		default:
		}
		lastHeight := getLastHeight()
		if h > lastHeight {
			break
		}
		bl.logger.Infof("load block data from[%d] to [%d], thread num [%d]", h, lastHeight, threadNum)
		dt := lastHeight - h + 1
		totalWork := uint64(workOneThread) * uint64(threadNum)
		if totalWork <= 0 {
			totalWork = dt
		}
		if dt < totalWork {
			threadNum = int32(dt) / workOneThread
			if threadNum == 0 {
				threadNum = 1
			}
			totalWork = dt
			bl.logger.Infof("actual number of threads: %d ", threadNum)
		}
		bl.logger.Infof("this round load block from[%d] to [%d], thread num [%d]", h, h+totalWork, threadNum)
		minHeight := bl.batchWork(h, totalWork, int(threadNum), stopCh)
		bl.logger.Infof("now load block to [%d]", minHeight)
		if minHeight > 0 {
			//将布隆过滤器的终态置为最新高度
			bl.loadBloom(minHeight, minHeight+1, stopCh)
			if err := bl.filter.Dump(); err != nil {
				bl.logger.Errorf("dump bloom filter error: %s", err)
			}
		}
		h = minHeight + 1
	}

	bl.logger.Infof("load block to height [%d]", h-1)
}

func (bl *blockBloomLoader) batchWork(h uint64, num uint64, threadNum int, stopCh <-chan struct{}) uint64 {
	var (
		wg           sync.WaitGroup
		i            = 0
		workers      = make([]*threadWorker, 0, threadNum)
		getMinHeight = func() uint64 {
			for i := 0; i < len(workers); i++ {
				worker := workers[i]
				if !worker.done() {
					return worker.process
				}
			}
			if len(workers) > 0 {
				return workers[len(workers)-1].process
			}
			return 0
		}
	)

	wg.Add(int(threadNum))
	per := uint64(num) / uint64(threadNum)
	for ; i < int(threadNum); i++ {
		worker := newThreadWorker(i, h+uint64(i)*per, h+uint64(i+1)*per, bl)
		workers = append(workers, worker)
		go func() {
			defer wg.Done()
			worker.run(stopCh)
		}()
	}
	if remain := uint64(num) % uint64(threadNum); remain > 0 {
		wg.Add(1)
		worker := newThreadWorker(i, h+uint64(i)*per, h+uint64(i)*per+remain, bl)
		workers = append(workers, worker)
		go func() {
			defer wg.Done()
			worker.run(stopCh)
		}()
	}
	wg.Wait()
	return getMinHeight()
}

func (bl *blockBloomLoader) loadBloom(from, to uint64, stopCh <-chan struct{}) uint64 {
	debugOpen := strings.ToLower(logLevel) == "debug"
	h := from
	for ; h < to; h++ {
		select {
		case <-stopCh:
			break
		default:
		}
		block, err := bl.GetBlockWithRWSets(h)
		if err != nil {
			bl.logger.Warnf("get block [%d] error: %s", h, err)
			h--
			time.Sleep(20 * time.Millisecond)
			continue
		}
		if debugOpen {
			count := 0
			for _, txRWSet := range block.TxRWSets {
				for _, write := range txRWSet.TxWrites {
					count++
					bl.logger.Debugf("write_key %s,  %s", string(write.ContractName), string(write.Key))
				}
			}
			bl.logger.Debugf("get block [%d], write_key.count [%d]", h, count)
		}
		bl.filter.AddBlock(block)
		if h%20 == 0 {
			bl.logger.Infof("load block to bloom filter [%d]", h)
		}
	}
	return h
}

func (bl *blockBloomLoader) Close() {
	for {
		if err := bl.filter.Dump(); err == nil {
			bl.logger.Info("dump successfully")
			break
		}
		// else if err != bloom.ErrDumpBusy {
		// 	bl.logger.Errorf("dump bloom data to file error: %s", err)
		// }
		time.Sleep(time.Second)
	}
	bl.blockFileDB.Close()
}

func openLevelDB(dbPath string) (*leveldb.DB, error) {
	return leveldb.OpenFile(dbPath, &opt.Options{
		ReadOnly: true,
	})
}

type threadWorker struct {
	idx     int
	from    uint64
	to      uint64
	process uint64
	bl      *blockBloomLoader
}

func newThreadWorker(idx int, from uint64, to uint64, bl *blockBloomLoader) *threadWorker {
	return &threadWorker{
		idx:  idx,
		from: from,
		to:   to,
		bl:   bl,
	}
}

func (w *threadWorker) run(stopCh <-chan struct{}) {
	debugOpen := strings.ToLower(logLevel) == "debug"
	bl, h := w.bl, w.from
	bl.logger.Infof("worker [%d] load block [%d, %d)", w.idx, w.from, w.to)
	for ; h < w.to; h++ {
		select {
		case <-stopCh:
			bl.logger.Infof("worker [%d] load block stop [%d]", w.idx, w.process)
			return
		default:
		}
		iter, err := bl.IterateBlockFrom(h)
		if err != nil {
			bl.logger.Warnf("get block [%d] error: %s", h, err)
			h--
			time.Sleep(20 * time.Millisecond)
			continue
		}
		bl.logger.Infof("worker [%d] get new iterator from [%d] to [%d]", w.idx, h, iter.LastHeight())
		for iter.HasNext() {
			select {
			case <-stopCh:
				bl.logger.Infof("worker [%d] load block stop [%d]", w.idx, w.process)
				return
			default:
			}
			block := iter.NextBlock()
			if debugOpen {
				count := 0
				for _, txRWSet := range block.TxRWSets {
					for _, write := range txRWSet.TxWrites {
						count++
						bl.logger.Debugf("worker [%d] write_key %s,  %s", w.idx, string(write.ContractName), string(write.Key))
					}
				}
				bl.logger.Debugf("worker [%d] get block [%d], write_key.count [%d]", w.idx, h, count)
			}
			bl.filter.AddBlock(block)
			h = block.Block.Header.BlockHeight
			w.process = h
			if h%20 == 0 {
				bl.logger.Infof("worker [%d] load block to bloom filter [%d]", w.idx, h)
			}
			// bl.logger.Infof("worker [%d] load block to bloom filter [%d]", w.idx, h)
			if h == w.to-1 {
				break
			}
		}
	}
	bl.logger.Infof("worker [%d] load block to [%d]", w.idx, w.process)
	// for ; h < w.to; h++ {
	// 	select {
	// 	case <-stopCh:
	// 		bl.logger.Infof("worker [%d] load block stop [%d]", w.idx, w.process)
	// 		return
	// 	default:
	// 	}
	// 	block, err := bl.GetBlockWithRWSets(h)
	// 	if err != nil {
	// 		bl.logger.Warnf("get block [%d] error: %s", h, err)
	// 		h--
	// 		time.Sleep(20 * time.Millisecond)
	// 		continue
	// 	}
	// 	if debugOpen {
	// 		count := 0
	// 		for _, txRWSet := range block.TxRWSets {
	// 			for _, write := range txRWSet.TxWrites {
	// 				count++
	// 				bl.logger.Debugf("worker [%d] write_key %s,  %s", w.idx, string(write.ContractName), string(write.Key))
	// 			}
	// 		}
	// 		bl.logger.Debugf("worker [%d] get block [%d], write_key.count [%d]", w.idx, h, count)
	// 	}
	// 	bl.filter.AddBlock(block)
	// 	w.process = h
	// 	if h%20 == 0 {
	// 		bl.logger.Infof("worker [%d] load block to bloom filter [%d]", w.idx, h)
	// 	}
	// 	time.Sleep(1 * time.Second)
	// }
	// bl.logger.Infof("worker [%d] load block to [%d]", w.idx, w.process)
}

func (w *threadWorker) done() bool {
	return w.process == (w.to - 1)
}

type blockIterator struct {
	from uint64
	data [][]byte
	idx  int
}

func (i *blockIterator) LastHeight() uint64 {
	return i.from + uint64(len(i.data)-1)
}

func (i *blockIterator) From() uint64 {
	return i.from
}

func (i *blockIterator) To() uint64 {

	return i.from + uint64(len(i.data)-1)
}

func (i *blockIterator) NextBlock() *storePb.BlockWithRWSet {
	block, err := serialization.DeserializeBlock(i.data[i.idx])
	if err != nil {
		fmt.Println("[ERROR] deserialize block[%d], error: %v", i.from+uint64(i.idx), err)
		panic(err)
		return nil
	}
	i.idx++
	return block
}
func (i *blockIterator) HasNext() bool {
	return i.idx < len(i.data)
}

// type blockIteratorCache struct {
// 	from  uint64
// 	to    uint64
// 	iters []*blockIterator
// 	idx   int
// 	len   int
// 	bl    *blockBloomLoader
// }

// func newBlockIteratorCache(cap int, from, to uint64, bl *blockBloomLoader) *blockIteratorCache {
// 	return &blockIteratorCache{
// 		from:  from,
// 		to:    to,
// 		iters: make([]*blockIterator, cap),
// 		len:   cap,
// 		bl:    bl,
// 	}
// }

// func (c *blockIteratorCache) createIteratorBackend() {

// }

// func (c *blockIteratorCache) getIterator(h uint64) *blockIterator {
// 	i := c.idx % c.len
// 	if iter := c.iters[i]; iter != nil {
// 		if h >= iter.From() && h <= iter.To() {
// 			c.idx++
// 			return iter
// 		}
// 	}
// 	iter, err := c.getFromDB(h)
// 	if err != nil {
// 		panic(err)
// 	}
// 	return iter
// }

// func (c *blockIteratorCache) getFromDB(h uint64) (*blockIterator, error) {
// 	return c.bl.IterateBlockFrom(h)
// }
